// * Project name:
//     MP3_Complete_Test
// * Copyright:
//     (c) MikroElektronika, 2009.
// * Description:
//     This project demonstrates communication with VS1011E mp3 codec.
//     Program reads one mp3 file from MMC and sends it to VS1011E for decoding
//     and playing.
//     MMC and MP3_SCI share Hardware SPI module.
//     MP3_SDI uses Software SPI communicaton.
//     Program progress can be monitored on UART.
//     Buttons RB7..RB4 are used for Left/Right Down/Up Volume control.
// * Test configuration:
//     MCU:             P18F4520
//     Dev.Board:       EasyPIC5
//     Oscillator:      HS-PLL, 08.000 MHz
//     Ext. Modules:    mikroElektronika's SmartMP3 board(hw rev 1.02) on PORTC and PORTD
//     SW:              mikroC PRO for PIC
// * NOTES:
//     - Make sure that MMC card is properly formatted (to FAT16 or just FAT)
//       before testing it on this example.
//     - Make sure that MMC card contains appropriate mp3 file.
//     - MMC FAT library is for PIC18 MCUs only.
//     - Turn ON the USART switches.
//     - SmartMP3 board connections:
//                MMC_CS  - RC0           DREQ  - RD0
//                MP3_CS  - RC1           BSYNC - RD1
//                MP3_RST - RC2           DCLK  - RD2
//                SCK     - RC3           SDATA - RD3
//                MISO    - RC4
//                MOSI    - RC5
//     - Set voltage level jumpers on SmartMP3 board into 5V position.
//     - Pull-down PORTB resistors and set Button jumper to Vcc position.

#include <built_in.h>
// Buttons for Left/Right Volume control
sbit BtnLeftDown            at RB7_bit;
sbit BtnLeftUp              at RB6_bit;
sbit BtnRightDown           at RB5_bit;
sbit BtnRightUp             at RB4_bit;
sbit BtnLeftDown_Direction  at TRISB7_bit;
sbit BtnLeftUp_Direction    at TRISB6_bit;
sbit BtnRightDown_Direction at TRISB5_bit;
sbit BtnRightUp_Direction   at TRISB4_bit;

// Smart MP3 board connections
sbit Mmc_Chip_Select           at RC0_bit;
sbit MP3_CS                    at RC1_bit;
sbit MP3_RST                   at RC2_bit;
sbit DREQ                      at RD0_bit;
sbit BSYNC                     at RD1_bit;
sbit DCLK                      at RD2_bit;
sbit SDATA                     at RD3_bit;
sbit Mmc_Chip_Select_Direction at TRISC0_bit;
sbit MP3_CS_Direction          at TRISC1_bit;
sbit MP3_RST_Direction         at TRISC2_bit;
sbit DREQ_Direction            at TRISD0_bit;
sbit BSYNC_Direction           at TRISD1_bit;
sbit DCLK_Direction            at TRISD2_bit;
sbit SDATA_Direction           at TRISD3_bit;

// VS1011E constants
const char WRITE_CODE  = 0x02;
const char READ_CODE   = 0x03;
const char MODE_ADDR   = 0x00;
const char CLOCKF_ADDR = 0x03;
const char VOL_ADDR    = 0x0B;

// global variables
char          filename[14] = "sound1.mp3";   // MP3 file name
unsigned long i, file_size;
char  data_buffer_32[32];
const BUFFER_SIZE = 512;
char  BufferLarge[BUFFER_SIZE];
char  volume_left, volume_right;

// Write text and goto new line CR and LF
void UART1_Write_Line (char *text) {
  UART1_Write_Text(text);
  UART1_Write(13);
  UART1_Write(10);
}

// Writes one byte to MP3 SDI
void SW_SPI_Write(char data_) {
  BSYNC = 1;                              // Set BSYNC before sending the first bit

  DCLK = 0;
  SDATA = data_;
  DCLK = 1;
  data_ >>= 1;                            // bitorder is LSB first

  DCLK = 0;
  SDATA = data_;
  DCLK = 1;
  data_ >>= 1;

  BSYNC = 0;                              // Clear BSYNC after sending the second bit

  DCLK = 0;
  SDATA = data_;
  DCLK = 1;
  data_ >>= 1;

  DCLK = 0;
  SDATA = data_;
  DCLK = 1;
  data_ >>= 1;

  DCLK = 0;
  SDATA = data_;
  DCLK = 1;
  data_ >>= 1;

  DCLK = 0;
  SDATA = data_;
  DCLK = 1;
  data_ >>= 1;

  DCLK = 0;
  SDATA = data_;
  DCLK = 1;
  data_ >>= 1;

  DCLK = 0;
  SDATA = data_;
  DCLK = 1;
  data_ >>= 1;

  DCLK = 0;
}

// Writes one word to MP3 SCI
void MP3_SCI_Write(char address, unsigned int data_in) {
  MP3_CS = 0;                             // select MP3 SCI
  SPI1_Write(WRITE_CODE);
  SPI1_Write(address);
  SPI1_Write(Hi(data_in));                // high byte
  SPI1_Write(Lo(data_in));                // low byte
  MP3_CS = 1;                             // deselect MP3 SCI
  while (DREQ == 0);                      // wait until DREQ becomes 1, see MP3 codec datasheet, Serial Protocol for SCI
}

// Reads words_count words from MP3 SCI
void MP3_SCI_Read(char start_address, char words_count, unsigned int *data_buffer) {
  unsigned int temp;

  MP3_CS = 0;                             // select MP3 SCI
  SPI1_Write(READ_CODE);
  SPI1_Write(start_address);

  while (words_count--) {                 // read words_count words byte per byte
    temp = SPI1_Read(0);
    temp <<= 8;
    temp += SPI1_Read(0);
    *(data_buffer++) = temp;
  }
  MP3_CS = 1;                             // deselect MP3 SCI
  while (DREQ == 0);                      // wait until DREQ becomes 1, see MP3 codec datasheet, Serial Protocol for SCI
}

// Write one byte to MP3 SDI
void MP3_SDI_Write(char data_) {
  while (DREQ == 0);                      // wait until DREQ becomes 1, see MP3 codec datasheet, Serial Protocol for SCI

  SW_SPI_Write(data_);
}

// Write 32 bytes to MP3 SDI
void MP3_SDI_Write_32(char *data_) {
  char i;

  while (DREQ == 0);                      // wait until DREQ becomes 1, see MP3 codec datasheet, Serial Protocol for SCI

  for (i=0; i<32; i++)
    SW_SPI_Write(data_[i]);
}

// Set clock
void Set_Clock(unsigned int clock_khz, char doubler) {
  clock_khz /= 2;                         // calculate value
  if (doubler) {
    clock_khz |= 0x8000;
  }
  MP3_SCI_Write(CLOCKF_ADDR, clock_khz);  // Write value to CLOCKF register
}

// Set volume
void Set_Volume(char left,char right) {
  unsigned int volume;

  volume = (left<<8) + right;             // calculate value
  MP3_SCI_Write(VOL_ADDR, volume);        // Write value to VOL register
}

// If volume Buttons are pressed update volume
void Check_Volume() {
  char old_volume_left, old_volume_right;

  old_volume_left  = volume_left;         // save volume values
  old_volume_right = volume_right;

  if (BtnLeftDown) {                      // RB7 pressed = LEFT VOLUME DOWN command
    if (volume_left < 255) {
      volume_left++;
    }
  }
  if (BtnLeftUp) {                        // RB6 pressed = LEFT VOLUME UP command
    if (volume_left > 0) {
      volume_left--;
    }
  }
  if (BtnRightDown) {                     // RB5 pressed = RIGHT VOLUME DOWN command
    if (volume_right < 255) {
      volume_right++;
    }
  }
  if (BtnRightUp) {                       // RB4 pressed = RIGHT VOLUME UP command
    if (volume_right > 0) {
      volume_right--;
    }
  }

  // if volume values are changed, set volume to new values
  if ( (volume_left != old_volume_left) || (volume_right != old_volume_right) ) {
    Set_Volume(volume_left,volume_right);
  }
}

// Software Reset
void Soft_Reset(char fill_zeros) {
  char i;

  MP3_SCI_Write(MODE_ADDR,0x0204);        // Write to MODE register
                                          //   set SM_RESET bit and SM_BITORD bit(bitorder is LSB first)
  Delay_us(2);                            // Required, see MP3 codec datasheet -> Software Reset
  while (DREQ == 0);                      // wait until DREQ becomes 1, see MP3 codec datasheet, Serial Protocol for SCI

  if (fill_zeros) {                       // feed 2048 zeros to the MP3 SDI bus:
    for (i=0; i<32; i++) {                // clear 32byte buffer
      data_buffer_32[i] = 0;
    }
    for (i=0; i < 2048/32; i++) {         // use 32byte buffer to clear MP3 buffer
      MP3_SDI_Write_32(data_buffer_32);
    }
  }
  else
    MP3_SDI_Write(0);                     // write zero once
}

// Initialize VS1011E
void MP3_init() {
  Soft_Reset(0);                          // SW Reset, do not feed zeros

  Set_Clock(25000,0);                     // Set clock to 25MHz, do not use clock doubler

  volume_left  = 50;                      // Set volume to initial value
  volume_right = 50;
  Set_Volume(volume_left,volume_right);
}

void Init() {
  ADCON1 |= 0x0F;                         // Configure AN pins as digital
  CMCON  |= 7;                            // Disable comparators

  BtnLeftDown_Direction  = 1;             // set pin direction as input
  BtnLeftUp_Direction    = 1;
  BtnRightDown_Direction = 1;
  BtnRightUp_Direction   = 1;
  BtnLeftDown  = 0;                       // set pin value to zero
  BtnLeftUp    = 0;
  BtnRightDown = 0;
  BtnRightUp   = 0;

  DCLK_Direction  = 0;                    // Set SW SPI pin directions to output
  SDATA_Direction = 0;
  DCLK  = 0;                              // Clear SW SPI SCK
  SDATA = 0;                              // Clear SW SPI SDO

  MP3_CS_Direction  = 0;                  // Configure MP3_CS as output
  MP3_CS            = 1;                  // Deselect MP3_CS
  MP3_RST_Direction = 0;                  // Configure MP3_RST as output
  MP3_RST           = 1;                  // Set MP3_RST pin

  DREQ_Direction  = 1;                    // Configure DREQ as input
  BSYNC_Direction = 0;                    // Configure BSYNC as output
  BSYNC           = 0;                    // Clear BSYNC
}

// main
void main() {
  Init();

  // Initialize USART for signaling
  UART1_Init(9600);
  Delay_100ms();

  UART1_Write_Line("Initializing SPI");
  SPI1_Init_Advanced(_SPI_MASTER_OSC_DIV64, _SPI_DATA_SAMPLE_MIDDLE, _SPI_CLK_IDLE_LOW, _SPI_LOW_2_HIGH);

  UART1_Write_Line("Initializing VS1011E decoder interface");
  MP3_init();

  UART1_Write_Line("Initializing MMC_FAT");
  if (Mmc_Fat_Init() == 0) {
    if (Mmc_Fat_Assign(&filename, 0) ) {
      UART1_Write_Line("File Assigned");
      Mmc_Fat_Reset(&file_size);          // Call Reset before file reading,
                                          //   procedure returns size of the file
      // send file blocks to MP3 SDI
      while (file_size > BUFFER_SIZE) {

        for (i=0; i<BUFFER_SIZE; i++) {
          Mmc_Fat_Read(BufferLarge + i);
        }

        for (i=0; i<BUFFER_SIZE/32; i++) {
          MP3_SDI_Write_32(BufferLarge + i*32);
        }

        file_size -= BUFFER_SIZE;

        Check_Volume();
      }

      // send the rest of the file to MP3 SDI
      for (i=0; i<file_size; i++) {
        Mmc_Fat_Read(BufferLarge + i);
      }

      for (i=0; i<file_size; i++) {
        MP3_SDI_Write(BufferLarge[i]);
      }

    }
    else {
      UART1_Write_Line("File not assigned");
    }

  }
  else {
    UART1_Write_Line("MMC FAT not initialized");
  }
}
